package com.yakindu.bridges.ea.example.codegen

import com.yakindu.bridges.ea.examples.runtime.codegen.EACodegen
import java.util.Date
import org.eclipse.uml2.uml.Class
import org.eclipse.uml2.uml.Element
import org.eclipse.uml2.uml.Enumeration
import org.eclipse.uml2.uml.Property
import org.eclipse.xtend2.lib.StringConcatenation

import static extension com.yakindu.bridges.ea.example.codegen.CodegenUtils.*

/**
 * This class contains several custom codegen templates that are evaluated immediately after save.
 * The resulting files are stored in sub-folder 'src-gen'.
 * 
 * Please feel free to change or extend the codegen templates, e.g. at the locations marked with XXX.
 * You could drag this editor to the right side of the editing area so that this and the generated files are shown side-by-side.
 * When modifying the templates, always keep an eye on the error log in case a template fails to be executed.
 * 
 * This class is written in <a href="http://xtend-lang.org">Xtend</a> and makes use of its template expressions (in triple quotes).
 * 
 * <b>Disclaimer:</b>
 * The YAKINDU EA-Bridge provides an extensive API to load, access, and partially also to modify EA models.
 * This example is only one (very simplified) demonstration to generate some code from EA models that are loaded with the YAKINDU EA-Bridge.
 * This example does neither have proper error handling, nor does scale in size and performance!
 * If you want to create a proper code generator, you should rather access the resource from your own plugins or in a JUnit Plugin test.
 */
class SampleCodegenTemplates {

	/* Enable this block (and organize imports via Ctrl + O) to change default generation target. */
//	@EACodegenFile
//	def IFile mapGeneratedFilePath(PackageableElement element, IFile genFile) {
//		val exampleProjectPath = genFile.fullPath.toString
//		val javaProjectPath = exampleProjectPath.replace(genFile.project.name, "LibraryInJava")
//		return ResourcesPlugin.workspace.root.getFile(new Path(javaProjectPath))
//	}

	/**
	 * Operations annotated with {@link EACodegen} are called for all model elements of the parameter's type.
	 * This operation, for example, is called for all UML classes in all eap-files within the enclosing project.
	 * It creates a java file with attributes via a <a href="https://www.eclipse.org/xtend/documentation/203_xtend_expressions.html#templates">template expression</a>.
	 */
	@EACodegen("java") // the file extension of the generated file; one file is generated per UML class.
	def generate(Class c) '''
«««		This is a comment inside a template expression (XXX: toggle comments with Ctrl + 7).
«««		XXX: A very simple way to test the code generator is to modify a comment, e.g. the one below.
		/**
		 * This file was generated by «this.class.simpleName» on «new Date».
		 */
		package «c.packageName»;

		«c.genJavadoc»
		«c.visibilityAsString»«IF c.abstract» abstract«ENDIF» class «c.name»«c.superClassAsString»«c.implementedInterfacesAsString» {

			«FOR a : c.ownedAttributes.filter[type !== null] SEPARATOR StringConcatenation.DEFAULT_LINE_DELIMITER»
				«a.genJavadoc»
«««				XXX: You could change code below to always create *private* fields and, for all public UML attributes, also generate public getters and setters.
«««				XXX: Another exercise is to respect the attribute's cardinality and generate {@link List}s for them.
				«a.visibilityAsString» «a.genType» «a.name»;
			«ENDFOR»
		}
	''' // if null is returned instead of a string or template expression, then no file is generated.

	/**
	 * This operation generates Java enumerations for UML enumerations.
	 */
	@EACodegen("java")
	def generate(Enumeration e) '''
		/**
		 * This file was generated by «this.class.simpleName» on «new Date».
		 */
		package «e.packageName»;

		«e.genJavadoc»
		public enum «e.name» {
			«e.ownedLiterals.map[name].join(", ")»
		}
	'''

	def genType(Property property) {
		if (property.type?.name.nullOrEmpty)
			return "void"
		// XXX: here you could change the code to return collections for multi-valued properties
		return property.type.name
	}

	/** A helper to generate UML comments as Javadoc, if available. */
	def genJavadoc(Element e) {
		val comments = e.ownedComments.filter[!body.nullOrEmpty]
		if (!comments.empty) '''
			/**
			 «FOR line : comments.map[body.split('\n').toList].flatten»
			 	* «line»
			 «ENDFOR»
			 */
		'''
	}

	/* XXX as exercise, you could add a new method that generates code for Interfaces in your model:
	 * 1. add an Interface to the ea-model, either using Enterprise Architect or the Eclipse UML editor
	 * 2. create a new operation: def generate(Interface i)
	 * 3. add annotation @EACodegen("java") to your operation
	 * 4. as operation body, write a template expression that creates a java interface
	 * 5. for all ownedOperations, generate their signatures 
	 * 6. save this file and observe the new java files being generated for all interfaces in your model
	 * */
}
