package com.yakindu.bridges.ea.example.validation;

import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.Property;

import com.yakindu.bridges.ea.examples.runtime.validation.EAValidation;
import com.yakindu.bridges.ea.examples.runtime.validation.Quickfix;
import com.yakindu.bridges.ea.examples.runtime.validation.Severity;

/**
 * This class contains several custom validation rules that are evaluated immediately after save.
 * 
 * Open the 'Problems' view (Window - Show View - Problems) for a list of all model violations.
 * Double click them to navigate directly to the violated model element.
 * 
 * Please feel free to change or extend the validation rules, e.g. at the locations marked with XXX.
 * When modifying rules, always keep an eye on the error log in case a validation rules fails to be executed.
 * In some cases, automatic re-validation fails. To trigger validation, clean project via 'Project' - 'Clean...'.
 * 
 * <b>Disclaimer:</b>
 * The YAKINDU EA-Bridge provides an extensive API to load, access, and partially also to modify EA models.
 * This example is only one (very simplified) demonstration to check constraints on EA models that are loaded with the YAKINDU EA-Bridge.
 * This example does neither have proper error handling, nor does scale in size and performance!
 * If you want proper model validation, you should rather access the resource from your own plugins.
 */
public class SampleValidationRules {

	/**
	 * This is a validation rule because it is annotated with {@link EAValidation}.
	 * It will be called for all UML elements in the entire model (so make sure that the validation logic is fast).
	 * 
	 * This validation rule reports a warning for all attributes without a type.
	 */
	// XXX change to 'error', save this file, observe validation results
	@EAValidation(severity = Severity.warning)
	public String validateAttributeTypes(Element element) {
		if (element instanceof Property) {
			if (((Property) element).getType() == null)
				// return message will be the validation message shown in Problems view
				return String.format("Attribute '%s' is missing a type!", ((NamedElement) element).getName());
		}
		return null; // return null if there is no violation
	}

	/** This is used in {@link #validateAbstractClass(Element)}. */
	private static final String ABSTRACT_CLASS_PREFIX = "Abstract";

	/**
	 * This rule reports all abstract classes whose names do not start with 'Abstract'.
	 *
	 * Furthermore, this validation rule refers to two quick fixes (by ID) that are defined below which can be used in
	 * the UI to automatically fix violations of this validation rule.
	 */
	@EAValidation(severity = Severity.warning, quickfix = { "rename", "toggleAbstract" })
	public String validateAbstractClass(Element element) {
		if (isAbstractClassWithoutPrefix(element, ABSTRACT_CLASS_PREFIX)) {
			return String.format("Abstract class '%s' should start with '%s'. Double click this in Problems view to navigate to the model element, right-click for possible quick fixes.",
					((Class) element).getName(), ABSTRACT_CLASS_PREFIX);
		}
		return null; // no violation
	}

	/** @returns whether or not the given element is an abstract class without the given prefix. */
	private boolean isAbstractClassWithoutPrefix(Element element, String prefix) {
		return element instanceof Class && ((Class) element).isAbstract() && !((Class) element).getName().startsWith(prefix);
	}

	/**
	 * This is a quick fix for the validation rule {@link #validateAbstractClass(Element)}. It should repair the model
	 * such that the validation rule is not violated any more. The return value indicates whether the model was changed.
	 * 
	 * Parameter <code>id</code> is used by {@link EAValidation} to refer to possible quick fixes. <code>label</code> is
	 * used in the UI.
	 */
	@Quickfix(id = "rename", label = "Add 'Abstract' prefix to class name")
	public boolean fixRenameAbstractClass(Element element) {
		if (isAbstractClassWithoutPrefix(element, ABSTRACT_CLASS_PREFIX)) {
			((Class) element).setName(ABSTRACT_CLASS_PREFIX + ((Class) element).getName());
			return true; // if the model was changed, 'true' must be returned
		}
		return false; // if the model was not changed, 'false' must be returned
	}

	/**
	 * Another quick fix for the same validation rule: this time, the 'abstract' flag of the class is toggled.
	 */
	@Quickfix(id = "toggleAbstract", label = "Toggle 'abstract' flag")
	public boolean fixToggleAbstract(Element element) {
		if (isAbstractClassWithoutPrefix(element, ABSTRACT_CLASS_PREFIX)) {
			((Class) element).setIsAbstract(false);
			return true; // if the model was changed, 'true' must be returned
		}
		return false; // if the model was not changed, 'false' must be returned
	}

	/* XXX as exercise, you could add a new validation rule that prohibits Use Cases in your model:
	 * 1. add a Use Case to the ea-model, either using Enterprise Architect or the Eclipse UML editor
	 * 2. create a new operation: public boolean validateNoUseCases(Element element)
	 * 3. add annotation @EAValidation to your operation and use code-completion (Ctrl + space) to set severity warning
	 * 4. in your operation body, return a message if the element is an instance of UseCase, return null otherwise
	 * 5. save this file and observe the validation results in the Problems view
	 * */

	/* XXX as follow-up exercise, you could add a quickfix for your validation rule that deletes use cases:
	 * 1. create a new operation: public boolean fixDeleteUseCase(Element element)
	 * 2. add annotation @Quickfix to your operation and set id and label
	 * 3. in your operation body, delete the element if it is a UseCase, e.g. with EcoreUtil.remove(EObject)
	 * 4. add quickfix by id to the annotation of your validation operation
	 * 5. save, wait for updated validation results, and run your quick-fix from the Problems view
	 * */
}
